/*--------------------------------------------------------------------*\
  File:      KThread.h
  Creator:   Matt Bogosian <mbogosian@usa.net>
  Copyright: (c)1998, Matt Bogosian. All rights reserved.
  Description: Header file describing the KThread abstract base class.
  ID:        $Id: KThread.h,v 1.30 1998/07/03 00:25:48 mattb Exp $
  Conventions:
      #defines - all uppercase letters with words separated by
          underscores.
          (E.G., #define MY_DEFINE 16).
      New data types (classes, structs, typedefs, etc.) - begin with
          an uppercase letter followed by lowercase words separated by
          uppercase letters.
          (E.G., typedef int MyTypedef;).
      Global constants (declared with const) - begin with "k_"
          followed by lowercase words separated by underscores.
          (E.G., const int k_my_constant(16);).
      Global variables - begin with "g_" followed by lowercase words
          separated by underscores.
          (E.G., int g_my_global;).
      Local variables - begin with a lowercase letter followed by
          lowercase words separated by underscores.
          (E.G., int my_local;).
      Argument variables - begin with "a_" followed by lowercase words
          separated by underscores.
          (E.G., ..., int a_my_argument, ...).
      Functions - begin with a lowercase letter followed by lowercase
          words separated by uppercase letters.
          (E.G., void myFunction(void);).
      Static member constants (declared with const) - begin with "mk_"
          followed by lowercase words separated by underscores.
          (E.G., static const int MyClass::mk_my_constant(16);).
      Static member variables - begin with "mg_" followed by lowercase
          words separated by underscores.
          (E.G., static int MyClass::mg_my_global(16);).
      Member variables - begin with "m_" followed by lowercase words
          separated by underscores.
          (E.G., int m_my_member;).
      Member Functions - begin with an uppercase letter followed by
          lowercase words separated by uppercase letters.
          (E.G., void MyClass::MyFunction(void);).
\*--------------------------------------------------------------------*/


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-=- Included Files -=-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#include "Benaphore.h"


#ifndef K_THREAD_H
#define K_THREAD_H


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-= Definitions, Enums, Typedefs, Consts =-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#ifndef DECL_SPEC

#if defined DLL_EXPORT_SYMBOLS
#define DECL_SPEC __declspec(dllexport)
#elif defined DLL_IMPORT_SYMBOLS
#define DECL_SPEC __declspec(dllimport)
#else
#define DECL_SPEC
#endif

#endif    // DECL_SPEC


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-= Structs, Classes =-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#ifdef DLL_EXPORT_SYMBOLS
#pragma export on
#endif

/*--------------------------------------------------------------------*\
  Class name:       KThread
  Inherits from:    none
  Defined in:       KThread.cpp
  New data members: private static const char * const mk_default_name
                        - the default name of the thread.
                    private thread m_thread - the kernel thread.
                    private Benaphore m_lock - the Benaphore to guard
                        access to the thread.
  Description: Abstract base class to implement a kernel thread of
      execution. This basically serves as a C++ wrapper to the kernel
      thread functions with a few extra goodies.
      
      To use, simply derive a class from KThread, define the Func()
      hook function:
      
        class MyThread : public KThread
        {
          // Define the Func() hook function here
        };
      
      Then instantiate it, make sure it's valid and Join() or Resume()
      it to start it on its way:
      
        MyThread *thread(new MyThread());
        
        if (thread != NULL)
        {
          if (thread->Status() == B_NO_ERROR)
          {
            // Wait for it to finish
            status_t ret_val;
            thread->Join(&ret_val);
          }
          else
          {
            delete thread;
          }
        }
      
      Note: unlike a BLooper, this thread object is not automatically
      deleted when its execution is finished. This is left up to
      derived classes to implement where needed.
\*--------------------------------------------------------------------*/

class DECL_SPEC KThread
{
	private:
	
		// Private static const data members
		static const char mk_default_name[B_OS_NAME_LENGTH];
	
	public:
	
		// Public static member functions
		
/*--------------------------------------------------------------------*\
  Function name: inline static Find
  Member of:     public KThread
  Defined in:    KThread.cpp
  Arguments:     const char a_name[B_OS_NAME_LENGTH] - the name of the
                     thread for which to search.
  Returns:       thread_id - the thread's ID on success, an error code
                     on failure (see below).
  Throws:        none
  Description: Function to attempt to find a thread with a given name.
      Possible return values are as follows:
      
      B_NAME_NOT_FOUND - the specified name was not found.
\*--------------------------------------------------------------------*/
		
		inline static thread_id Find(const char a_name[B_OS_NAME_LENGTH])
		{
			return find_thread(a_name);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline static KillTeam
  Member of:     public KThread
  Arguments:     const team_id a_team - the team to kill.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to kill a given team. Possible return values
      are as follows:
      
      B_NO_ERROR - success.
      B_BAD_TEAM_ID - the specified team was not valid.
\*--------------------------------------------------------------------*/
		
		inline static status_t KillTeam(const team_id a_team)
		{
			return kill_team(a_team);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline static TeamInfo
  Member of:     public KThread
  Arguments:     const team_id a_team - the team about which to get
                     the info.
                 team_info &a_info - a placeholder for the team's
                     info.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to get information about a given team.
      Possible return values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_TEAM_ID - the specified team was not valid.
\*--------------------------------------------------------------------*/
		
		inline static status_t TeamInfo(const team_id a_team, team_info &a_info)
		{
			return get_team_info(a_team, &a_info);
		}
		
		// Public member functions
		
/*--------------------------------------------------------------------*\
  Function name: KThread
  Member of:     public KThread
  Defined in:    KThread.cpp
  Arguments:     const char a_name[B_OS_NAME_LENGTH] - the name of the
                     thread. Default: mk_default_name.
                 const int32 a_priority - the priority of the thread.
                     Default: B_NORMAL_PRIORITY.
                 status_t * const a_err - a placeholder for the error
                     encountered during construction (see below).
                     Default: NULL.
  Throws:        none
  Description: Default class constructor. You must check the value of
      a_err or the return value of Status() after invoking the
      constructor to see if the object is usable. See the Status()
      function (below) for possible error values.
\*--------------------------------------------------------------------*/
		
		KThread(const char a_name[B_OS_NAME_LENGTH] = mk_default_name, const int32 a_priority = B_NORMAL_PRIORITY, status_t * const a_err = NULL);
		
/*--------------------------------------------------------------------*\
  Function name: virtual ~KThread
  Member of:     public KThread
  Defined in:    KThread.cpp
  Arguments:     none
  Throws:        none
  Description: Class destructor.
\*--------------------------------------------------------------------*/
		
		virtual ~KThread(void);
		
/*--------------------------------------------------------------------*\
  Function name: inline Info const
  Member of:     public KThread
  Arguments:     thread_info &a_info - a placeholder for the thread's
                     info.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to return the info about the thread. This
      should be used for informative purposes only. Note: you should
      call Lock() before accessing this function. Possible return
      values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_VALUE - this thread is no longer valid.
\*--------------------------------------------------------------------*/
		
		inline status_t Info(thread_info &a_info) const
		{
			return get_thread_info(m_thread, &a_info);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline Join
  Member of:     public KThread
  Arguments:     status_t &a_ret_val - a placeholder for the return
                     value from this thread when it completes
                     execution.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to join the calling thread with this thread.
      This function blocks until this thread has finished its
      execution for that reason, you should not call this function
      from within the thread itself. One way to insure proper
      execution is to Lock() this thread, Suspend() it, then Unlock()
      it and call this function:
      
        if (thread->Lock() == B_NO_ERROR)
        {
          thread->Suspend();
          // Tell the thread it should quit here if necessary
          thread->Unlock();
          thread->Join(...);
        }
      
      This only works (of course) if no one is trying to kill the
      thread in the mean time. Note: this function Lock()s the thread,
      so you should not Lock() it yourself before calling it. Possible
      return values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_THREAD_ID - this thread is no longer valid.
      B_INTERRUPTED - the thread has halted execution prematurely
          (i.e., via exit_thread(), kill_thread() or kill_team()).
\*--------------------------------------------------------------------*/
		
		inline status_t Join(status_t &a_ret_val)
		{
			if (Status() != B_NO_ERROR)
			{
				return B_BAD_THREAD_ID;
			}
			
			return wait_for_thread(m_thread, &a_ret_val);
		}
		
/*--------------------------------------------------------------------*\
  Function name: Kill
  Member of:     public KThread
  Defined in:    KThread.cpp
  Arguments:     none
  Returns:       none
  Throws:        none
  Description: Function to kill this thread. Note: this function
      Lock()s the thread, so you should not Lock() it yourself before
      calling it.
\*--------------------------------------------------------------------*/
		
		void Kill(void);
		
/*--------------------------------------------------------------------*\
  Function name: inline Lock
  Member of:     public KThread
  Arguments:     const uint32 a_flags - the flags passed to
                     acquire_sem_etc(). Default: 0.
                 const bigtime_t a_timeout - the timeout in
                     microseconds. Default: B_INFINITE_TIMEOUT.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to lock the thread. A successful call to
      Lock() (i.e., on which returns B_NO_ERROR) should always be
      mirrored by exactly one call to Unlock(). Note: as an safety
      implementation detail, in the worst case, this function might
      pause twice as long as the indicated timeout before returning
      a B_TIMED_OUT result. In the best case, it will only take as
      long as the indicated timeout. If your locking is dependant on
      precision timing, do not use this class. Possible return values
      are as follows:
      
      B_NO_ERROR - success.
      B_BAD_SEM_ID - this thread is not valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the thread's
          Benaphore was not available.
      B_TIMED_OUT - the timeout was reached and the thread's Benaphore
          was not acquired.
\*--------------------------------------------------------------------*/
		
		inline status_t Lock(const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT)
		{
			return m_lock.Lock(a_flags, a_timeout);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline Rename
  Member of:     public KThread
  Arguments:     const char * const a_name - the new name of the
                     thread.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to rename the thread. Note: you should call
  Lock() before accessing this function. Possible return values are as
  follows:
      
      B_NO_ERROR - success.
      B_BAD_THREAD_ID - this thread is no longer valid.
\*--------------------------------------------------------------------*/
		
		inline status_t Rename(const char * const a_name)
		{
			return rename_thread(m_thread, a_name);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline Resume
  Member of:     public KThread
  Arguments:     none
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to resume execution of the thread. Note: you
      should call Lock() before accessing this function. Possible
      return values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_THREAD_ID - this thread is no longer valid.
      B_BAD_THREAD_STATE - this thread is not suspended.
\*--------------------------------------------------------------------*/
		
		inline status_t Resume(void)
		{
			if (Status() != B_NO_ERROR)
			{
				return B_BAD_THREAD_ID;
			}
			
			return resume_thread(m_thread);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline Send
  Member of:     public KThread
  Arguments:     const int32 a_msg_code - the message code.
                 const void * const a_msg - the message. Default:
                     NULL.
                 const size_t a_msg_size - the length (in bytes) of
                     the message. Default: 0.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to send a message to the thread. This function
      will block until the message is sent or an error occurs. Note:
      you should call Lock() before accessing this function. Be
      careful with thread messages. It's very easy to get into a
      double-locking-like race condition here. To avoid this, make
      sure that the thread is actually reading the messages, and that
      reading them doesn't depend on Lock()ing the thread. Possible
      return values are as follows:
      
      B_NO_ERROR - success.
      B_NO_MEMORY - there was not enough memory to copy the buffer.
      B_BAD_THREAD_ID - this thread is no longer valid.
      B_INTERRUPTED - this function interrupted by a signal.
\*--------------------------------------------------------------------*/
		
		inline status_t Send(const int32 a_msg_code, const void * const a_msg = NULL, const size_t a_msg_len = 0)
		{
			return send_data(m_thread, a_msg_code, a_msg, a_msg_len);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline SetPriority
  Member of:     public KThread
  Arguments:     const int32 a_priority - the new priority.
  Returns:       status_t - the previous priority on success, an error
                     code on failure (see below).
  Throws:        none
  Description: Function to set the priority of execution of this
      thread. Note: you should call Lock() before accessing this
      function. Possible return values are as follows:
      
      B_BAD_THREAD_ID - this thread is no longer valid.
\*--------------------------------------------------------------------*/
		
		inline status_t SetPriority(const int32 a_priority)
		{
			return set_thread_priority(m_thread, a_priority);
		}
		
/*--------------------------------------------------------------------*\
  Function name: virtual Status const
  Member of:     public KThread
  Defined in:    KThread.cpp
  Arguments:     none
  Returns:       status_t - B_NO_ERROR if the object is valid, another
                     error code if not (see below).
  Throws:        none
  Description: Function to check if the object is valid. This function
      must be called after creating the object to insure that its
      instantiation was successful. This function should probably not
      be used once the thread is in motion. Classes who wish to
      override this function, should always call this function first
      to insure that the underlying structure is valid. Possible
      return values are as follows:
      
      B_NO_ERROR - the object was created successfully.
      B_BAD_THREAD_ID - this thread is no longer valid.
      B_NO_MEMORY - on construction, there was not enough memory
          available to create the object.
      B_NO_MORE_SEMS - on construction, there were no available
          semaphores to create the object's Benaphore.
      B_NO_MORE_THREADS - on construction, there were no available
          threads for the object.
\*--------------------------------------------------------------------*/
		
		virtual status_t Status(void) const;
		
/*--------------------------------------------------------------------*\
  Function name: inline Suspend
  Member of:     public KThread
  Arguments:     none
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to suspend the execution of this thread. Note:
      you should call Lock() before accessing this function. Possible
      return values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_THREAD_ID - this thread is no longer valid.
\*--------------------------------------------------------------------*/
		
		inline status_t Suspend(void)
		{
			return suspend_thread(m_thread);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline Unlock
  Member of:     public KThread
  Arguments:     const uint32 a_flags - the flags passed to
                     release_sem_etc(). Default: 0.
  Returns:       none
  Throws:        none
  Description: Function to unlock the thread.
\*--------------------------------------------------------------------*/
		
		inline void Unlock(const uint32 a_flags = 0)
		{
			m_lock.Unlock(a_flags);
		}
	
	protected:
	
		// Protected static member functions
		
/*--------------------------------------------------------------------*\
  Function name: static Receive
  Member of:     protected KThread
  Defined in:    KThread.cpp
  Arguments:     int32 &a_msg_code - a placeholder for the message
                     code.
                 thread_id * const a_thread - a placeholder for the
                     thread from whence the message came. Default:
                     NULL.
                 void * const a_buf - the buffer in which to store the
                     message. Default: NULL.
                 const size_t a_buf_size - the size (in bytes) of the
                     buffer. Default: 0.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to receive a message sent to this thead via
      either the Send() member function or the send_data() kernel
      function. This function will block until a message is received
      or an error occurs (it's usually a good idea to call
      MessageWaiting() before calling this function unless you're
      anticipating a message). If this function fails, the message
      code and thread placeholders remain untouched. Possible return
      values are as follows:
      
      B_NO_ERROR - success.
      B_NO_MEMORY - there was not enough memory to create the thread
          to be returned.
      B_BAD_THREAD_ID - this thread is no longer valid.
      B_INTERRUPTED - this function interrupted by a signal.
\*--------------------------------------------------------------------*/
		
		static status_t Receive(int32 &a_msg_code, thread_id * const a_thread = NULL, void * const a_buf = NULL, const size_t a_buf_size = 0);
		
/*--------------------------------------------------------------------*\
  Function name: inline static Snooze
  Member of:     protected KThread
  Arguments:     const bigtime_t a_time - the length of time (in
                     microseconds) to pause execution.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to pause the thread's execution for a given
      length of time. Possible return values are as follows:
      
      B_NO_ERROR - success.
      B_ERROR - the given length of time was invalid (i.e., less than
          0).
      B_INTERRUPTED - this function interrupted by a signal.
\*--------------------------------------------------------------------*/
		
		inline static status_t Snooze(const bigtime_t a_time)
		{
			return snooze(a_time);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline static SnoozeUntil
  Member of:     protected KThread
  Arguments:     const bigtime_t a_time - the time at which the thread
                     should resume execution.
                 const int a_time_base - the time base which will be
                     used to calculate when the time will be reached.
                     Default: B_SYSTEM_TIMEBASE.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to pause the thread's execution until a given
      time. Possible return values are as follows:
      
      B_NO_ERROR - success.
      B_ERROR - the given length of time was invalid (i.e., less than
          0).
      B_INTERRUPTED - this function interrupted by a signal.
\*--------------------------------------------------------------------*/
		
		inline static status_t SnoozeUntil(const bigtime_t a_time, const int a_time_base = B_SYSTEM_TIMEBASE)
		{
			return snooze_until(a_time, a_time_base);
		}
		
		// Protected member functions
		
/*--------------------------------------------------------------------*\
  Function name: inline Exit
  Member of:     protected KThread
  Arguments:     const status_t a_ret_val - the return value of this
                     thread.
  Returns:       none
  Throws:        none
  Description: Function to exit execution of this thread. Note: this
      function Lock()s the thread, so you should not Lock() it
      yourself before calling it.
\*--------------------------------------------------------------------*/
		
		inline void Exit(const status_t a_ret_val)
		{
			// Invalidate the thread
			if (Lock() == B_NO_ERROR)
			{
				m_thread = B_BAD_THREAD_ID;
				Unlock();
			}
			
			exit_thread(a_ret_val);
		}
		
/*--------------------------------------------------------------------*\
  Function name: virtual Func
  Member of:     protected KThread
  Arguments:     none
  Returns:       status_t - return value for the thread.
  Throws:        none
  Description: Hook function called during the thread's execution.
\*--------------------------------------------------------------------*/
		
		virtual status_t Func(void) = NULL;
		
/*--------------------------------------------------------------------*\
  Function name: inline MessageWaiting
  Member of:     protected KThread
  Arguments:     none
  Returns:       bool - true if there is a message waiting, false if
                     not.
  Throws:        none
  Description: Function to check to see if the thread has a message
      waiting to be received (e.g., with Receive()).
\*--------------------------------------------------------------------*/
		
		inline bool MessageWaiting(void)
		{
			return has_data(m_thread);
		}
	
	private:
	
		// Private data members
		thread_id m_thread;
		Benaphore m_lock;
		
		// Private static member functions
		
/*--------------------------------------------------------------------*\
  Function name: inline static ThreadFunc
  Member of:     private KThread
  Arguments:     void * const a_data - a pointer to the instance.
  Returns:       status_t - return value for the thread.
  Throws:        none
  Description: Function to call the Func() member function from within
      the thread's execution.
\*--------------------------------------------------------------------*/
		
		inline static status_t ThreadFunc(void * const a_data)
		{
			KThread *thread(static_cast<KThread *>(a_data));
			status_t ret_val(thread->Func());
			
			// Invalidate the thread
			if (thread->Lock() == B_NO_ERROR)
			{
				thread->m_thread = B_BAD_THREAD_ID;
				thread->Unlock();
			}
			
			return ret_val;
		}
		
		// Private prohibitted member functions
		KThread(const KThread &a_obj);
		KThread &operator=(const KThread &a_obj);
};

#ifdef DLL_EXPORT_SYMBOLS
#pragma export off
#endif


#endif    // K_THREAD_H
